#include <cassert>
#include <iostream>
#include <functional>
#include <map>
#include <memory>
#include <string>
#include <vector>
//#include <type_traits>
#include "tiny_ecs.hpp"

// All data relevant to the shape and motion of entities in this example
struct Motion {
	float position = 0.0f;
	float velocity = 1.0f;
};

// A simplified ECSRegistry
class ECSRegistry
{
public:
	// Manually created list of all components this game has
	ComponentContainer<Motion> motions;
};
ECSRegistry registry;

// The return type of behaviour tree processing
enum class BTState {
    Running,
    Success,
    Failure
};

std::string toString(BTState s)
{
	switch (s)
	{
	case BTState::Running:   return "Running";
	case BTState::Success:   return "Success";
	case BTState::Failure: return "Failure";
	default:      return "[Unknown BTState]";
	}
}

// The base class representing any node in our behaviour tree
class BTNode {
public:
	virtual void init(Entity e) {};

	virtual BTState process(Entity e) = 0;
};

// A composite node that loops through all children and exits when one fails
class BTRunPair : public BTNode {
private:
	int m_index;
	BTNode* m_children[2];

public:
	BTRunPair(BTNode* c0, BTNode* c1)
		: m_index(0) {
		m_children[0] = c0;
		m_children[1] = c1;
	}

	void init(Entity e) override
	{
		m_index = 0;
		// initialize the first child
		const auto& child = m_children[m_index];
		child->init(e);
	}

	BTState process(Entity e) override {
		if (m_index >= 2)
			return BTState::Success;

		// process current child
		BTState state = m_children[m_index]->process(e);

		// select a new active child and initialize its internal state
		if (state == BTState::Success) {
			++m_index;
			if (m_index >= 2) {
				return BTState::Success;
			}
			else {
				m_children[m_index]->init(e);
				return BTState::Running;
			}
		}
		else {
			return state;
		}
	}
};

// A general decorator with lambda condition
class BTIfCondition : public BTNode
{
public:
	BTIfCondition(BTNode* child, std::function<bool(Entity)> condition)
		: m_child(child), m_condition(condition){
	}

	virtual void init(Entity e) override {
		m_child->init(e);
	}

	virtual BTState process(Entity e) override {
		if (m_condition(e))
			return m_child->process(e);
		else
			return BTState::Success;
	}

private:
	BTNode* m_child;
	std::function<bool(Entity)> m_condition;
};

// A general decorator with lambda condition
class BTIfConditionHardCoded : public BTNode
{
public:
	BTIfConditionHardCoded(BTNode* child)
		: m_child(child) {
	}

	virtual void init(Entity e) override {
		m_child->init(e);
	}

	virtual BTState process(Entity e) override {
		if (registry.motions.has(e))
			return m_child->process(e);
		else
			return BTState::Success;
	}
private:
	BTNode* m_child;
};

// Leaf node examples
class RunNSteps : public BTNode {
public:
    RunNSteps(int steps) noexcept
        : m_targetSteps(steps), m_stepsRemaining(0) {
    }

private:
	void init(Entity e) override {
		m_stepsRemaining = m_targetSteps;
	}

	BTState process(Entity e) override {
		// update internal state
		--m_stepsRemaining;

		// modify world
        auto& motion = registry.motions.get(e);
        motion.position += motion.velocity;

		// return progress
		if (m_stepsRemaining > 0) {
			return BTState::Running;
        } else {
			return BTState::Success;
        }
	}

private:
    int m_targetSteps;
	int m_stepsRemaining;
};

class TurnAround : public BTNode {
private:
	void init(Entity e) override {
    }

	BTState process(Entity e) override {
		// modify world
		auto& vel = registry.motions.get(e).velocity;
		vel = -vel;

		// return progress
		return BTState::Success;
	}
};

/////////////////////////////////////////
// Demo entry point
int main(int argc, char* argv[]) {
	// Notes
	// run and turn leaf nodes
	RunNSteps run3(3);
	TurnAround turn;
	RunNSteps run1(1);

	// conditional turn sub-tree
	std::function<bool(Entity)> condition = [](Entity e) {return registry.motions.get(e).velocity < 0; };
	BTIfCondition turn_right(&turn, condition);

	// entire tree, starting at the root
	BTRunPair root_run_and_turn(&run3, &turn);
	BTRunPair root_run_and_return(&root_run_and_turn, &run1);

	BTRunPair root_turn_right_and_run(&turn_right, &root_run_and_turn);

	// apply trees on horse and zebra
	Entity horse, zebra;
	registry.motions.emplace(horse);
	//registry.motions.emplace(zebra);
	root_run_and_return.init(horse);
	//root_turn_and_continue_right.init(zebra);

    std::cout << "Start\n";
	std::cout << "  velocity = " << registry.motions.get(horse).velocity << '\n';
	std::cout << "  position = " << registry.motions.get(horse).position << '\n';

	for (int i = 0; i < 100; i++) {
		std::cout << "step " << i << '\n';
		BTState state = root_run_and_return.process(horse);

		std::cout << "  horse: \n";
		std::cout << "  velocity = " << registry.motions.get(horse).velocity << '\n';
		std::cout << "  position = " << registry.motions.get(horse).position << '\n';

		if (registry.motions.has(zebra))
		{
			BTState state2 = root_turn_right_and_run.process(zebra);
			std::cout << "  zebra: \n";
			std::cout << "  velocity = " << registry.motions.get(zebra).velocity << '\n';
			std::cout << "  position = " << registry.motions.get(zebra).position << '\n';
		}

		std::cout << "  state = " << toString(state)<< '\n';
		if (state != BTState::Running) {
			break;
        }
	}

	return EXIT_SUCCESS;
}
